/* main.c

    Written by Clément NOTIN for Erasme Centre Multimédia
    Summer 2007
*/
#include <stdio.h>
#include <string.h>  //memset()
#include <pwd.h>
#include <unistd.h> //close(), sleep()
#include <fcntl.h>   //FD settings
#include <signal.h>
#include <stdlib.h> //exit()
#include <confuse.h>

#include "rpn.h"
#include "serial.h"
#include "daemonize.h"
#include "osc.h"

#define TRUE 1
#define FALSE 0

int fd; //global serial port file descriptor
int signalNum=0; //will contain signal num

//global var initalized with the conf settings
char *OSC_IP_ADDRESS=NULL;
char *OSC_ADDRESS=NULL;
char *OSC_PORT=NULL;
char *ARDUINO_DEVICE=NULL;
char *USER=NULL;
char *GROUP=NULL;
int BAUDRATE=0;
int NB_SENSORS=0;
int DAEMONIZE=0;
int OSC_ENABLED=0;
int REFRESH_TIME=0;
char *STATUS_FILE=NULL;
char *PID_FILE=NULL;



//##########################################################
//dummy callback for the signal catching, sets signalNum to a non-zero value in order to stop the main while loop
void signalHandler(int signalId)
{
    if(signalId == SIGUSR1)
    { writeport("x"); }
    else
    { signalNum=signalId; }
}    


//################################# MAIN ########################
int main(int argc, char **argv) {
    //classical signal handler    
    struct sigaction sa;
    memset(&sa, 0, sizeof sa);
    sa.sa_handler = signalHandler;
    sa.sa_flags=0;
    sigemptyset(&(sa.sa_mask));
    if(sigaction(SIGINT, &sa, NULL) != 0)
        perror("Sigaction problem ");
    if(sigaction(SIGHUP, &sa, NULL) != 0)
        perror("Sigaction problem ");
    if(sigaction(SIGUSR1, &sa, NULL) != 0)
        perror("Sigaction problem ");       
        
    //configuration part, take a look at libconfuse documentation
    //here we define the default value for each setting
    cfg_opt_t sensorOptions[] =
    {
        CFG_STR("name", "Unnamed", CFGF_NONE),
        CFG_STR("rpn", "X", CFGF_NONE),
        CFG_BOOL("memorize", FALSE, CFGF_NONE),
        CFG_BOOL("memorize_state", TRUE, CFGF_NONE),
        CFG_END()
    };   
    cfg_opt_t mainOptions[] = //define the param names and default values read from the conf file
    {
        CFG_STR("osc_ip_address", "127.0.0.1", CFGF_NONE),
        CFG_STR("osc_address", "/sensors/%s", CFGF_NONE),
        CFG_STR("osc_port", "1234", CFGF_NONE),
        CFG_STR("arduino_device", "/dev/ttyUSB0", CFGF_NONE),
        CFG_STR("user", "sensorsd", CFGF_NONE),
        CFG_STR("group", "dialout", CFGF_NONE),
        CFG_INT("baudrate", 9600, CFGF_NONE),
        CFG_INT("refresh_time", 10, CFGF_NONE),
        CFG_BOOL("daemonize", TRUE, CFGF_NONE),
        CFG_BOOL("osc_enabled", TRUE, CFGF_NONE),
        CFG_STR("status_file", "/var/local/sensorsd/status", CFGF_NONE),
        CFG_STR("pid_file", "/var/run/sensorsd.pid", CFGF_NONE),
        CFG_SEC("sensor", sensorOptions, CFGF_TITLE | CFGF_MULTI),
        CFG_END()
    };
    cfg_t *mainConfig, *sensorConfig;
    mainConfig = cfg_init(mainOptions, CFGF_NOCASE);

    //we can specify a config file in the command line
    if(argc > 1)
    {
        printf("Loading config file %s\n", argv[1]);
        if(cfg_parse(mainConfig, argv[1]) == CFG_PARSE_ERROR)
            return 1;
    }
    else //if no file was specified, we load the default one
    {
        if(cfg_parse(mainConfig, "/etc/sensorsd.conf") == CFG_PARSE_ERROR)
            return 1;
    }
    
    //read the values from the file
    OSC_IP_ADDRESS=cfg_getstr(mainConfig, "osc_ip_address");
    OSC_ADDRESS=cfg_getstr(mainConfig, "osc_address");
    OSC_PORT=cfg_getstr(mainConfig, "osc_port");
    ARDUINO_DEVICE=cfg_getstr(mainConfig, "arduino_device");
    USER=cfg_getstr(mainConfig, "user");
    GROUP=cfg_getstr(mainConfig, "group");
    BAUDRATE=cfg_getint(mainConfig, "baudrate");
    NB_SENSORS=cfg_size(mainConfig, "sensor");
    STATUS_FILE=cfg_getstr(mainConfig, "status_file");
    PID_FILE=cfg_getstr(mainConfig, "pid_file");
    DAEMONIZE=cfg_getbool(mainConfig, "daemonize");
    OSC_ENABLED=cfg_getbool(mainConfig, "osc_enabled");
    REFRESH_TIME=cfg_getint(mainConfig, "refresh_time");
    
    //contains 4 infos about each sensor
    typedef struct sensorsOpt sensorsOpt;
    struct sensorsOpt
    {
        const char *name; //sensor's name
        const char *sCode; //sensor's code as it is written in the file
        char cCode; //sensor's code as it will be sent to the arduino
        char *rpn; //RPN expression
        int memorize; //do we must memorize ?
        int memorize_state; //which state do we must memorise ?
        int memorize_locked; //set to TRUE if we have found the matching value, if TRUE : the value won't be updated until SIGHUP
        int value;
    };
    sensorsOpt sensors[NB_SENSORS];
    unsigned int i;
    for(i=0; i < NB_SENSORS; i++) //init the infos about each sensor
    {
        sensorConfig=cfg_getnsec(mainConfig, "sensor", i); //load the section
        sensors[i].name = cfg_getstr(sensorConfig, "name");
        sensors[i].sCode = cfg_title(sensorConfig); //the code is the title of the section
        
        if(sensors[i].sCode[0] >= '0' && sensors[i].sCode[0] <= '5') //we must convert the char to the right number : e.g. '0' -> '\x00', '4' -> '\x04'
        { sensors[i].cCode = atoi(&sensors[i].sCode[0]); }
        else
        { sensors[i].cCode = sensors[i].sCode[0]; } //transform a 2 bytes string into a 1 byte char

        sensors[i].rpn = cfg_getstr(sensorConfig, "rpn");
        sensors[i].memorize = cfg_getbool(sensorConfig, "memorize");
        sensors[i].memorize_state = cfg_getbool(sensorConfig, "memorize_state");
        sensors[i].memorize_locked = FALSE;
        sensors[i].value = 0;
    }


    if (DAEMONIZE == TRUE)
        daemonize();

	fd = open(ARDUINO_DEVICE, O_RDWR | O_NOCTTY | O_NDELAY);
	
	if (fd == -1) {
		perror("open_port: Unable to open ARDUINO_DEVICE ");
		exit(1);
	} else {
		fcntl(fd, F_SETFL, 0);
	}
	
	//sensorStatus will contain the state or value of each sensor
	FILE *sensorsStatus;
    if ( (sensorsStatus = fopen(STATUS_FILE, "w")) == NULL)
        { perror("Unable to open sensorsStatus file "); return 1; }

	initport();

    int value=0; //will contains the value read from the serial


    while(signalNum != SIGINT) //loops until we received SIGINT
    {
        rewind(sensorsStatus); //clear the status file

        //loop on each sensors
        for(i=0; i < NB_SENSORS; i++)
        {
            if(!writeport(&(sensors[i].cCode))) //ask the arduino the value
            {
        		close(fd);
        		return 1;
        	}

            value=rpn(sensors[i].rpn, readport()); //compute the RPN expression
            printf("real value for %s = %d\n", sensors[i].name, value);
            
            if(signalNum == SIGHUP) //release the locks
            { sensors[i].memorize_locked = FALSE; }
            
            if( sensors[i].memorize && !sensors[i].memorize_locked ) //can we memorize ?
            {
                if( (value == sensors[i].value) && (value == sensors[i].memorize_state) ) //the value is matching => lock
                { 
                    printf("memorizing %d for %s!!!\n", value, sensors[i].name);
                    sensors[i].memorize_locked = TRUE;
                }
                if( !sensors[i].memorize_locked )
                { sensors[i].value = value; }
            }
            if( !sensors[i].memorize ) //not a memorized sensors => just set the value
            { sensors[i].value = value; }
            
            if(OSC_ENABLED == TRUE)
            { oscSend(sensors[i].name, sensors[i].value); } //tell the world what the sensor is saying
            if(DAEMONIZE == FALSE)
            { printf("%s,%s,%d\n", sensors[i].sCode, sensors[i].name, sensors[i].value); } //used mainly for debugging
            //write to the file in a grep friendly format what the sensor is saying
            fprintf(sensorsStatus, "%s,%s,%d\n", sensors[i].sCode, sensors[i].name, sensors[i].value);
            fflush(sensorsStatus); //be sure the datas are really written in the file
        }
        if(DAEMONIZE == FALSE)
        { printf("\n\n-----------------\n\n"); }
        
        if(signalNum == SIGHUP) //all sensors are unlocked, now signalNum is useless
        { signalNum = 0; }

        sleep(REFRESH_TIME); //have a break !
	}

	close(fd);
	rewind(sensorsStatus); //clear the status file so we know that the program is NOT running
	fclose(sensorsStatus);
	cfg_free(mainConfig); //we don't want the memory to leak !
	return 0;
}
